/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * ai_int.c                                                  *
 * part of checkers project, contains functions alpha-beta   *
 * function making the computer play international checkers  *
 * Miroslav Tynovsky 03/2004                                 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <stdlib.h>
#include <assert.h>
#include "rules_int.h"
#include "ai_int.h"

#define PAWN_VALUE 1
#define KING_VALUE 4

#define is_out_of_board(x, y) ((x)>=BOARDSIZE || (x)<0 || (y)>=BOARDSIZE || (y)<0)
#define correct_coords(x, y) (!is_out_of_board((x), (y)))

int evaluate_position(Position *position);
int alpha_beta(Position *position, Occured_position *occured_positions,
	           Move last_move, int depth, int alpha, int beta);
void sort_moves(Moves moves, Position *position);

Move choose_best_move(Position *position, int skill)
{
	int i;
	Moves moves;
	Move best_move;
	int current_ev, best_ev= -INFINITY;
	int best_index= 0;

	moves= get_all_moves(position);
	assert(moves.count>0);

	for(i=0; i<moves.count; i++){
		do_move(position, moves.list[i]);
		current_ev= -alpha_beta(position, NULL, moves.list[i], 2*skill, -INFINITY, INFINITY);
//		current_ev= -evaluate_position(position);
		if(current_ev >= best_ev){
			best_index= i;
			best_ev= current_ev;
		}
		undo_move(position, moves.list[i]);
	}
	best_move= moves.list[best_index];
	free(moves.list);
	assert(correct_coords(best_move.start.x, best_move.start.y));
	return best_move;
}

/* alpha_beta() returns evaluation of given position computed 
 * by alpha-beta algorithm
 */
int alpha_beta(Position *position, Occured_position *occured_positions,
	           Move last_move, int depth, int alpha, int beta)
{ 
	Moves moves;
	int i, ab;
	Position pos, pos2; //for debugging

	switch(get_game_state(position, occured_positions, &last_move)){
	case WHITE_WINS:
		return position->onmove * INFINITY;
	case BLACK_WINS:
		return -position->onmove * INFINITY;
	case DRAW_RP:
	case DRAW:
		return 0;
	}
	if(depth==0) return evaluate_position(position);
	moves= get_all_moves(position);
	sort_moves(moves, position);
	for(i=0; i<moves.count; i++){ 
		pos= *position; //debug
		do_move(position, moves.list[i]);
		pos2= *position; //debug
		ab= -alpha_beta(position, occured_positions, moves.list[i], depth-1,-beta,-alpha);
		assert(eq_positions(*position, pos2)); //debug
		undo_move(position, moves.list[i]); 
		assert(eq_positions(*position, pos)); //debug
		if(ab > alpha){
			alpha= ab;
			if(ab >= beta){
				free(moves.list);
				return beta;
			}
		}
	}
	free(moves.list);
	return alpha;
}

/* evaluate_position() returns static evaluation of given position

int evaluate_position(Position *position)
{
	int x, y, evaluation= 0;

	for(y=0; y<BOARDSIZE; y++){
		for(x=0; x<BOARDSIZE; x++){
			switch(position->placement[x][y]){
			case BLACK_PAWN: evaluation-= PAWN_VALUE;  break; 
			case WHITE_PAWN: evaluation+= PAWN_VALUE;  break; 
			case BLACK_KING: evaluation-= KING_VALUE;  break; 
			case WHITE_KING: evaluation+= KING_VALUE;  break; 
			}
		}
	}
	evaluation*= position->onmove;
	return evaluation;
}
 */

 
const int zero[50] = {
 0,  0,  0,  0,  0,
   0,  0,  0,  0,  0,
 0,  0,  0,  0,  0,
   0,  0,  0,  0,  0,
 0,  0,  0,  0,  0,
   0,  0,  0,  0,  0,
 0,  0,  0,  0,  0,
   0,  0,  0,  0,  0,
 0,  0,  0,  0,  0,
   0,  0,  0,  0,  0
};

const int whiteMan[50] = {
  8,       8,       9,       8,       8,
      9,       9,       9,       9,       9, 
 10,      10,      10,      10,      10,
     11,      11,      11,      11,      11,
 12,      14,      14,      12,      12,
     13,      13,      14,      15,      13, 
 14,      14,      14,      14,      14,
     15,      15,      15,      15,      15,
 16,      16,      16,      16,      16,
     17,      17,      17,      17,      17 
};

const int blackMan[50] = {
-17,     -17,     -17,     -17,     -17,
    -16,     -16,     -16,     -16,     -16, 
-15,     -15,     -15,     -15,     -15,
    -14,     -14,     -14,     -14,     -14,
-13,     -15,     -14,     -13,     -13, 
    -12,     -12,     -14,     -14,     -12,
-11,     -11,     -11,     -11,     -11,
    -10,     -10,     -10,     -10,     -10,
 -9,      -9,      -9,      -9,      -9, 
     -8,      -8,      -9,      -8,      -8
};

const int whiteKing[50] = {
     20,      20,      20,      20,      20,
 20,      20,      20,      20,      20,
     20,      20,      20,      20,      20,
 20,      20,      22,      22,      20,
     20,      23,      23,      20,      20,
 20,      20,      23,      23,      20,
     20,      22,      22,      20,      20,
 20,      20,      20,      20,      20,
     20,      20,      20,      20,      20,
 20,      20,      20,      20,      20
};

const int blackKing[50] = {
    -20,     -20,     -20,     -20,     -20,
-20,     -20,     -20,     -20,     -20,
    -20,     -20,     -20,     -20,     -20,
-20,     -20,     -22,     -22,     -20,
    -20,     -23,     -23,     -20,     -20,
-20,     -20,     -23,     -23,     -20,
    -20,     -22,     -22,     -20,     -20,
-20,     -20,     -20,     -20,     -20,
    -20,     -20,     -20,     -20,     -20,
-20,     -20,     -20,     -20,     -20
};

const int* value[5] = {
    blackKing, blackMan, zero, whiteMan, whiteKing
};

int evaluate_position(const Position *position)
{
    int x, y, index, e = 0;

    for(x=0; x<BOARDSIZE; x++)
		for(y=0; y<BOARDSIZE; y++)  {
			index= x/2 + y*((BOARDSIZE+1) / 2);
			e+= value[position->placement[x][y]+2][index];
		}
    return position->onmove*e;
} 


/* sorts moves according to their evaluations. best move first.
 */
void sort_moves(Moves moves, Position *position)
{
	int i, j, k, eval;
	Move move;
	Position pos; //debug
	int *evaluations= (int *)malloc(moves.count * sizeof(int));	

	for(i=0; i<moves.count; i++)  {
		pos= *position; //debug
		do_move(position, moves.list[i]);
		evaluations[i]= evaluate_position(position);
		undo_move(position, moves.list[i]);
		assert(eq_positions(*position, pos)); //debug
	}
	for(i=0; i<moves.count-1; i++)  {
		k= i;
		for(j=i+1; j<moves.count; j++)
			if(evaluations[j] < evaluations[k]) k= j;
		if(k > i){
			/* swap the values: */
			eval= evaluations[k];
			evaluations[k]= evaluations[i];
			evaluations[i]= eval;
			/* swap the moves: */
			move= moves.list[k];
			moves.list[k]= moves.list[i];
			moves.list[i]= move;
		}
	}
	free(evaluations);
} /* sort_moves() */

